// Copyright 2016 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

package larch64

import (
	"math"

	"cmd/compile/internal/gc"
	"cmd/compile/internal/logopt"
	"cmd/compile/internal/ssa"
	"cmd/compile/internal/types"
	"cmd/internal/obj"
	"cmd/internal/obj/larch64"
)

// isFPreg reports whether r is an FP register
func isFPreg(r int16) bool {
	return larch64.REG_F0 <= r && r <= larch64.REG_F31
}

// loadByType returns the load instruction of the given type.
func loadByType(t *types.Type, r int16) obj.As {
	if isFPreg(r) {
		if t.Size() == 4 { // float32 or int32
			return larch64.AMOVF
		} else { // float64 or int64
			return larch64.AMOVD
		}
	} else {
		switch t.Size() {
		case 1:
			if t.IsSigned() {
				return larch64.AMOVB
			} else {
				return larch64.AMOVBU
			}
		case 2:
			if t.IsSigned() {
				return larch64.AMOVH
			} else {
				return larch64.AMOVHU
			}
		case 4:
			if t.IsSigned() {
				return larch64.AMOVW
			} else {
				return larch64.AMOVWU
			}
		case 8:
			return larch64.AMOVV
		}
	}
	panic("bad load type")
}

// storeByType returns the store instruction of the given type.
func storeByType(t *types.Type, r int16) obj.As {
	if isFPreg(r) {
		if t.Size() == 4 { // float32 or int32
			return larch64.AMOVF
		} else { // float64 or int64
			return larch64.AMOVD
		}
	} else {
		switch t.Size() {
		case 1:
			return larch64.AMOVB
		case 2:
			return larch64.AMOVH
		case 4:
			return larch64.AMOVW
		case 8:
			return larch64.AMOVV
		}
	}
	panic("bad store type")
}

func ssaGenValue(s *gc.SSAGenState, v *ssa.Value) {
	switch v.Op {
	case ssa.OpCopy, ssa.OpLARCH64MOVVreg:
		if v.Type.IsMemory() {
			return
		}
		x := v.Args[0].Reg()
		y := v.Reg()
		if x == y {
			return
		}
		as := larch64.AMOVV
		if isFPreg(x) && isFPreg(y) {
			as = larch64.AMOVD
		}
		p := s.Prog(as)
		p.From.Type = obj.TYPE_REG
		p.From.Reg = x
		p.To.Type = obj.TYPE_REG
		p.To.Reg = y
	case ssa.OpLARCH64MOVVnop:
		if v.Reg() != v.Args[0].Reg() {
			v.Fatalf("input[0] and output not in same register %s", v.LongString())
		}
		// nothing to do
	case ssa.OpLoadReg:
		if v.Type.IsFlags() {
			v.Fatalf("load flags not implemented: %v", v.LongString())
			return
		}
		r := v.Reg()
		p := s.Prog(loadByType(v.Type, r))
		gc.AddrAuto(&p.From, v.Args[0])
		p.To.Type = obj.TYPE_REG
		p.To.Reg = r
	case ssa.OpStoreReg:
		if v.Type.IsFlags() {
			v.Fatalf("store flags not implemented: %v", v.LongString())
			return
		}
		r := v.Args[0].Reg()
		p := s.Prog(storeByType(v.Type, r))
		p.From.Type = obj.TYPE_REG
		p.From.Reg = r
		gc.AddrAuto(&p.To, v)
	case ssa.OpLARCH64ADDV,
		ssa.OpLARCH64SUBV,
		ssa.OpLARCH64AND,
		ssa.OpLARCH64OR,
		ssa.OpLARCH64XOR,
		ssa.OpLARCH64NOR,
		ssa.OpLARCH64SLLV,
		ssa.OpLARCH64SRLV,
		ssa.OpLARCH64SRAV,
		ssa.OpLARCH64ADDF,
		ssa.OpLARCH64ADDD,
		ssa.OpLARCH64SUBF,
		ssa.OpLARCH64SUBD,
		ssa.OpLARCH64MULF,
		ssa.OpLARCH64MULD,
		ssa.OpLARCH64DIVF,
		ssa.OpLARCH64DIVD:
		p := s.Prog(v.Op.Asm())
		p.From.Type = obj.TYPE_REG
		p.From.Reg = v.Args[1].Reg()
		p.Reg = v.Args[0].Reg()
		p.To.Type = obj.TYPE_REG
		p.To.Reg = v.Reg()
	case ssa.OpLARCH64SGT,
		ssa.OpLARCH64SGTU:
		p := s.Prog(v.Op.Asm())
		p.From.Type = obj.TYPE_REG
		p.From.Reg = v.Args[0].Reg()
		p.Reg = v.Args[1].Reg()
		p.To.Type = obj.TYPE_REG
		p.To.Reg = v.Reg()
	case ssa.OpLARCH64ADDVconst,
		ssa.OpLARCH64SUBVconst,
		ssa.OpLARCH64ANDconst,
		ssa.OpLARCH64ORconst,
		ssa.OpLARCH64XORconst,
		ssa.OpLARCH64NORconst,
		ssa.OpLARCH64SLLVconst,
		ssa.OpLARCH64SRLVconst,
		ssa.OpLARCH64SRAVconst,
		ssa.OpLARCH64SGTconst,
		ssa.OpLARCH64SGTUconst:
		p := s.Prog(v.Op.Asm())
		p.From.Type = obj.TYPE_CONST
		p.From.Offset = v.AuxInt
		p.Reg = v.Args[0].Reg()
		p.To.Type = obj.TYPE_REG
		p.To.Reg = v.Reg()
	case ssa.OpLARCH64MULV:
		p := s.Prog(larch64.AMULV)
		p.From.Type = obj.TYPE_REG
		p.From.Reg = v.Args[1].Reg()
		p.Reg = v.Args[0].Reg()
		p.To.Type = obj.TYPE_REG
		p.To.Reg = v.Reg1()
		p1 := s.Prog(larch64.AMULHV)
		p1.From.Type = obj.TYPE_REG
		p1.From.Reg = v.Args[1].Reg()
		p1.Reg = v.Args[0].Reg()
		p1.To.Type = obj.TYPE_REG
		p1.To.Reg = v.Reg0()
	case ssa.OpLARCH64MULVU:
		p := s.Prog(larch64.AMULV)
		p.From.Type = obj.TYPE_REG
		p.From.Reg = v.Args[1].Reg()
		p.Reg = v.Args[0].Reg()
		p.To.Type = obj.TYPE_REG
		p.To.Reg = v.Reg1()
		p1 := s.Prog(larch64.AMULHVU)
		p1.From.Type = obj.TYPE_REG
		p1.From.Reg = v.Args[1].Reg()
		p1.Reg = v.Args[0].Reg()
		p1.To.Type = obj.TYPE_REG
		p1.To.Reg = v.Reg0()
	case ssa.OpLARCH64DIVV:
		p := s.Prog(larch64.ADIVV)
		p.From.Type = obj.TYPE_REG
		p.From.Reg = v.Args[1].Reg()
		p.Reg = v.Args[0].Reg()
		p.To.Type = obj.TYPE_REG
		p.To.Reg = v.Reg1()
		p1 := s.Prog(larch64.AREMV)
		p1.From.Type = obj.TYPE_REG
		p1.From.Reg = v.Args[1].Reg()
		p1.Reg = v.Args[0].Reg()
		p1.To.Type = obj.TYPE_REG
		p1.To.Reg = v.Reg0()
	case ssa.OpLARCH64DIVVU:
		p := s.Prog(larch64.ADIVVU)
		p.From.Type = obj.TYPE_REG
		p.From.Reg = v.Args[1].Reg()
		p.Reg = v.Args[0].Reg()
		p.To.Type = obj.TYPE_REG
		p.To.Reg = v.Reg1()
		p1 := s.Prog(larch64.AREMVU)
		p1.From.Type = obj.TYPE_REG
		p1.From.Reg = v.Args[1].Reg()
		p1.Reg = v.Args[0].Reg()
		p1.To.Type = obj.TYPE_REG
		p1.To.Reg = v.Reg0()
	case ssa.OpLARCH64MOVVconst:
		r := v.Reg()
		p := s.Prog(v.Op.Asm())
		p.From.Type = obj.TYPE_CONST
		p.From.Offset = v.AuxInt
		p.To.Type = obj.TYPE_REG
		p.To.Reg = r
		if isFPreg(r) {
			// cannot move into FP or special registers, use TMP as intermediate
			p.To.Reg = larch64.REGTMP
			p = s.Prog(larch64.AMOVV)
			p.From.Type = obj.TYPE_REG
			p.From.Reg = larch64.REGTMP
			p.To.Type = obj.TYPE_REG
			p.To.Reg = r
		}
	case ssa.OpLARCH64MOVFconst,
		ssa.OpLARCH64MOVDconst:
		p := s.Prog(v.Op.Asm())
		p.From.Type = obj.TYPE_FCONST
		p.From.Val = math.Float64frombits(uint64(v.AuxInt))
		p.To.Type = obj.TYPE_REG
		p.To.Reg = v.Reg()
	case ssa.OpLARCH64CMPEQF,
		ssa.OpLARCH64CMPEQD,
		ssa.OpLARCH64CMPGEF,
		ssa.OpLARCH64CMPGED,
		ssa.OpLARCH64CMPGTF,
		ssa.OpLARCH64CMPGTD:
		p := s.Prog(v.Op.Asm())
		p.From.Type = obj.TYPE_REG
		p.From.Reg = v.Args[0].Reg()
		p.Reg = v.Args[1].Reg()
	case ssa.OpLARCH64MOVVaddr:
		p := s.Prog(larch64.AMOVV)
		p.From.Type = obj.TYPE_ADDR
		p.From.Reg = v.Args[0].Reg()
		var wantreg string
		// MOVV $sym+off(base), R
		// the assembler expands it as the following:
		// - base is SP: add constant offset to SP (R29)
		//               when constant is large, tmp register (R23) may be used
		// - base is SB: load external address with relocation
		switch v.Aux.(type) {
		default:
			v.Fatalf("aux is of unknown type %T", v.Aux)
		case *obj.LSym:
			wantreg = "SB"
			gc.AddAux(&p.From, v)
		case *gc.Node:
			wantreg = "SP"
			gc.AddAux(&p.From, v)
		case nil:
			// No sym, just MOVV $off(SP), R
			wantreg = "SP"
			p.From.Offset = v.AuxInt
		}
		if reg := v.Args[0].RegName(); reg != wantreg {
			v.Fatalf("bad reg %s for symbol type %T, want %s", reg, v.Aux, wantreg)
		}
		p.To.Type = obj.TYPE_REG
		p.To.Reg = v.Reg()
	case ssa.OpLARCH64MOVBload,
		ssa.OpLARCH64MOVBUload,
		ssa.OpLARCH64MOVHload,
		ssa.OpLARCH64MOVHUload,
		ssa.OpLARCH64MOVWload,
		ssa.OpLARCH64MOVWUload,
		ssa.OpLARCH64MOVVload,
		ssa.OpLARCH64MOVFload,
		ssa.OpLARCH64MOVDload:
		p := s.Prog(v.Op.Asm())
		p.From.Type = obj.TYPE_MEM
		p.From.Reg = v.Args[0].Reg()
		gc.AddAux(&p.From, v)
		p.To.Type = obj.TYPE_REG
		p.To.Reg = v.Reg()
	case ssa.OpLARCH64MOVBstore,
		ssa.OpLARCH64MOVHstore,
		ssa.OpLARCH64MOVWstore,
		ssa.OpLARCH64MOVVstore,
		ssa.OpLARCH64MOVFstore,
		ssa.OpLARCH64MOVDstore:
		p := s.Prog(v.Op.Asm())
		p.From.Type = obj.TYPE_REG
		p.From.Reg = v.Args[1].Reg()
		p.To.Type = obj.TYPE_MEM
		p.To.Reg = v.Args[0].Reg()
		gc.AddAux(&p.To, v)
	case ssa.OpLARCH64MOVBstorezero,
		ssa.OpLARCH64MOVHstorezero,
		ssa.OpLARCH64MOVWstorezero,
		ssa.OpLARCH64MOVVstorezero:
		p := s.Prog(v.Op.Asm())
		p.From.Type = obj.TYPE_REG
		p.From.Reg = larch64.REGZERO
		p.To.Type = obj.TYPE_MEM
		p.To.Reg = v.Args[0].Reg()
		gc.AddAux(&p.To, v)
	case ssa.OpLARCH64MOVBreg,
		ssa.OpLARCH64MOVBUreg,
		ssa.OpLARCH64MOVHreg,
		ssa.OpLARCH64MOVHUreg,
		ssa.OpLARCH64MOVWreg,
		ssa.OpLARCH64MOVWUreg:
		a := v.Args[0]
		for a.Op == ssa.OpCopy || a.Op == ssa.OpLARCH64MOVVreg {
			a = a.Args[0]
		}
		if a.Op == ssa.OpLoadReg {
			t := a.Type
			switch {
			case v.Op == ssa.OpLARCH64MOVBreg && t.Size() == 1 && t.IsSigned(),
				v.Op == ssa.OpLARCH64MOVBUreg && t.Size() == 1 && !t.IsSigned(),
				v.Op == ssa.OpLARCH64MOVHreg && t.Size() == 2 && t.IsSigned(),
				v.Op == ssa.OpLARCH64MOVHUreg && t.Size() == 2 && !t.IsSigned(),
				v.Op == ssa.OpLARCH64MOVWreg && t.Size() == 4 && t.IsSigned(),
				v.Op == ssa.OpLARCH64MOVWUreg && t.Size() == 4 && !t.IsSigned():
				// arg is a proper-typed load, already zero/sign-extended, don't extend again
				if v.Reg() == v.Args[0].Reg() {
					return
				}
				p := s.Prog(larch64.AMOVV)
				p.From.Type = obj.TYPE_REG
				p.From.Reg = v.Args[0].Reg()
				p.To.Type = obj.TYPE_REG
				p.To.Reg = v.Reg()
				return
			default:
			}
		}
		fallthrough
	case ssa.OpLARCH64MOVWF,
		ssa.OpLARCH64MOVWD,
		ssa.OpLARCH64TRUNCFW,
		ssa.OpLARCH64TRUNCDW,
		ssa.OpLARCH64MOVVF,
		ssa.OpLARCH64MOVVD,
		ssa.OpLARCH64TRUNCFV,
		ssa.OpLARCH64TRUNCDV,
		ssa.OpLARCH64MOVFD,
		ssa.OpLARCH64MOVDF,
		ssa.OpLARCH64NEGF,
		ssa.OpLARCH64NEGD,
		ssa.OpLARCH64SQRTD:
		p := s.Prog(v.Op.Asm())
		p.From.Type = obj.TYPE_REG
		p.From.Reg = v.Args[0].Reg()
		p.To.Type = obj.TYPE_REG
		p.To.Reg = v.Reg()
	case ssa.OpLARCH64NEGV:
		// SUB from REGZERO
		p := s.Prog(larch64.ASUBVU)
		p.From.Type = obj.TYPE_REG
		p.From.Reg = v.Args[0].Reg()
		p.Reg = larch64.REGZERO
		p.To.Type = obj.TYPE_REG
		p.To.Reg = v.Reg()
	case ssa.OpLARCH64DUFFZERO:
		// runtime.duffzero expects start address - 8 in R1
		p := s.Prog(larch64.ASUBVU)
		p.From.Type = obj.TYPE_CONST
		p.From.Offset = 8
		p.Reg = v.Args[0].Reg()
		p.To.Type = obj.TYPE_REG
		p.To.Reg = larch64.REG_R19
		p = s.Prog(obj.ADUFFZERO)
		p.To.Type = obj.TYPE_MEM
		p.To.Name = obj.NAME_EXTERN
		p.To.Sym = gc.Duffzero
		p.To.Offset = v.AuxInt
	case ssa.OpLARCH64LoweredZero:
		// SUBV	$8, R1
		// MOVV	R0, 8(R1)
		// ADDV	$8, R1
		// BNE	Rarg1, R1, -2(PC)
		// arg1 is the address of the last element to zero
		var sz int64
		var mov obj.As
		switch {
		case v.AuxInt%8 == 0:
			sz = 8
			mov = larch64.AMOVV
		case v.AuxInt%4 == 0:
			sz = 4
			mov = larch64.AMOVW
		case v.AuxInt%2 == 0:
			sz = 2
			mov = larch64.AMOVH
		default:
			sz = 1
			mov = larch64.AMOVB
		}
		p := s.Prog(larch64.ASUBVU)
		p.From.Type = obj.TYPE_CONST
		p.From.Offset = sz
		p.To.Type = obj.TYPE_REG
		p.To.Reg = larch64.REG_R19
		p2 := s.Prog(mov)
		p2.From.Type = obj.TYPE_REG
		p2.From.Reg = larch64.REGZERO
		p2.To.Type = obj.TYPE_MEM
		p2.To.Reg = larch64.REG_R19
		p2.To.Offset = sz
		p3 := s.Prog(larch64.AADDVU)
		p3.From.Type = obj.TYPE_CONST
		p3.From.Offset = sz
		p3.To.Type = obj.TYPE_REG
		p3.To.Reg = larch64.REG_R19
		p4 := s.Prog(larch64.ABNE)
		p4.From.Type = obj.TYPE_REG
		p4.From.Reg = v.Args[1].Reg()
		p4.Reg = larch64.REG_R19
		p4.To.Type = obj.TYPE_BRANCH
		gc.Patch(p4, p2)
	case ssa.OpLARCH64DUFFCOPY:
		p := s.Prog(obj.ADUFFCOPY)
		p.To.Type = obj.TYPE_MEM
		p.To.Name = obj.NAME_EXTERN
		p.To.Sym = gc.Duffcopy
		p.To.Offset = v.AuxInt
	case ssa.OpLARCH64LoweredMove:
		// SUBV	$8, R1
		// MOVV	8(R1), Rtmp
		// MOVV	Rtmp, (R2)
		// ADDV	$8, R1
		// ADDV	$8, R2
		// BNE	Rarg2, R1, -4(PC)
		// arg2 is the address of the last element of src
		var sz int64
		var mov obj.As
		switch {
		case v.AuxInt%8 == 0:
			sz = 8
			mov = larch64.AMOVV
		case v.AuxInt%4 == 0:
			sz = 4
			mov = larch64.AMOVW
		case v.AuxInt%2 == 0:
			sz = 2
			mov = larch64.AMOVH
		default:
			sz = 1
			mov = larch64.AMOVB
		}
		p := s.Prog(larch64.ASUBVU)
		p.From.Type = obj.TYPE_CONST
		p.From.Offset = sz
		p.To.Type = obj.TYPE_REG
		p.To.Reg = larch64.REG_R19
		p2 := s.Prog(mov)
		p2.From.Type = obj.TYPE_MEM
		p2.From.Reg = larch64.REG_R19
		p2.From.Offset = sz
		p2.To.Type = obj.TYPE_REG
		p2.To.Reg = larch64.REGTMP
		p3 := s.Prog(mov)
		p3.From.Type = obj.TYPE_REG
		p3.From.Reg = larch64.REGTMP
		p3.To.Type = obj.TYPE_MEM
		p3.To.Reg = larch64.REG_R4
		p4 := s.Prog(larch64.AADDVU)
		p4.From.Type = obj.TYPE_CONST
		p4.From.Offset = sz
		p4.To.Type = obj.TYPE_REG
		p4.To.Reg = larch64.REG_R19
		p5 := s.Prog(larch64.AADDVU)
		p5.From.Type = obj.TYPE_CONST
		p5.From.Offset = sz
		p5.To.Type = obj.TYPE_REG
		p5.To.Reg = larch64.REG_R4
		p6 := s.Prog(larch64.ABNE)
		p6.From.Type = obj.TYPE_REG
		p6.From.Reg = v.Args[2].Reg()
		p6.Reg = larch64.REG_R19
		p6.To.Type = obj.TYPE_BRANCH
		gc.Patch(p6, p2)
	case ssa.OpLARCH64CALLstatic, ssa.OpLARCH64CALLclosure, ssa.OpLARCH64CALLinter:
		s.Call(v)
	case ssa.OpLARCH64LoweredWB:
		p := s.Prog(obj.ACALL)
		p.To.Type = obj.TYPE_MEM
		p.To.Name = obj.NAME_EXTERN
		p.To.Sym = v.Aux.(*obj.LSym)
	case ssa.OpLARCH64LoweredPanicBoundsA, ssa.OpLARCH64LoweredPanicBoundsB, ssa.OpLARCH64LoweredPanicBoundsC:
		p := s.Prog(obj.ACALL)
		p.To.Type = obj.TYPE_MEM
		p.To.Name = obj.NAME_EXTERN
		p.To.Sym = gc.BoundsCheckFunc[v.AuxInt]
		s.UseArgs(16) // space used in callee args area by assembly stubs
	case ssa.OpLARCH64LoweredAtomicLoad8, ssa.OpLARCH64LoweredAtomicLoad32, ssa.OpLARCH64LoweredAtomicLoad64:
		as := larch64.AMOVV
		switch v.Op {
		case ssa.OpLARCH64LoweredAtomicLoad8:
			as = larch64.AMOVB
		case ssa.OpLARCH64LoweredAtomicLoad32:
			as = larch64.AMOVW
		}
		s.Prog(larch64.ASYNC)
		p := s.Prog(as)
		p.From.Type = obj.TYPE_MEM
		p.From.Reg = v.Args[0].Reg()
		p.To.Type = obj.TYPE_REG
		p.To.Reg = v.Reg0()
		s.Prog(larch64.ASYNC)
	case ssa.OpLARCH64LoweredAtomicStore8, ssa.OpLARCH64LoweredAtomicStore32, ssa.OpLARCH64LoweredAtomicStore64:
		as := larch64.AMOVV
		switch v.Op {
		case ssa.OpLARCH64LoweredAtomicStore8:
			as = larch64.AMOVB
		case ssa.OpLARCH64LoweredAtomicStore32:
			as = larch64.AMOVW
		}
		s.Prog(larch64.ASYNC)
		p := s.Prog(as)
		p.From.Type = obj.TYPE_REG
		p.From.Reg = v.Args[1].Reg()
		p.To.Type = obj.TYPE_MEM
		p.To.Reg = v.Args[0].Reg()
		s.Prog(larch64.ASYNC)
	case ssa.OpLARCH64LoweredAtomicStorezero32, ssa.OpLARCH64LoweredAtomicStorezero64:
		as := larch64.AMOVV
		if v.Op == ssa.OpLARCH64LoweredAtomicStorezero32 {
			as = larch64.AMOVW
		}
		s.Prog(larch64.ASYNC)
		p := s.Prog(as)
		p.From.Type = obj.TYPE_REG
		p.From.Reg = larch64.REGZERO
		p.To.Type = obj.TYPE_MEM
		p.To.Reg = v.Args[0].Reg()
		s.Prog(larch64.ASYNC)
	case ssa.OpLARCH64LoweredAtomicExchange32, ssa.OpLARCH64LoweredAtomicExchange64:
		// SYNC
		// MOVV	Rarg1, Rtmp
		// LL	(Rarg0), Rout
		// SC	Rtmp, (Rarg0)
		// BEQ	Rtmp, -3(PC)
		// SYNC
		ll := larch64.ALLV
		sc := larch64.ASCV
		if v.Op == ssa.OpLARCH64LoweredAtomicExchange32 {
			ll = larch64.ALL
			sc = larch64.ASC
		}
		s.Prog(larch64.ASYNC)
		p := s.Prog(larch64.AMOVV)
		p.From.Type = obj.TYPE_REG
		p.From.Reg = v.Args[1].Reg()
		p.To.Type = obj.TYPE_REG
		p.To.Reg = larch64.REGTMP
		p1 := s.Prog(ll)
		p1.From.Type = obj.TYPE_MEM
		p1.From.Reg = v.Args[0].Reg()
		p1.To.Type = obj.TYPE_REG
		p1.To.Reg = v.Reg0()
		p2 := s.Prog(sc)
		p2.From.Type = obj.TYPE_REG
		p2.From.Reg = larch64.REGTMP
		p2.To.Type = obj.TYPE_MEM
		p2.To.Reg = v.Args[0].Reg()
		p3 := s.Prog(larch64.ABEQ)
		p3.From.Type = obj.TYPE_REG
		p3.From.Reg = larch64.REGTMP
		p3.To.Type = obj.TYPE_BRANCH
		gc.Patch(p3, p)
		s.Prog(larch64.ASYNC)
	case ssa.OpLARCH64LoweredAtomicAdd32, ssa.OpLARCH64LoweredAtomicAdd64:
		// SYNC
		// LL	(Rarg0), Rout
		// ADDV Rarg1, Rout, Rtmp
		// SC	Rtmp, (Rarg0)
		// BEQ	Rtmp, -3(PC)
		// SYNC
		// ADDV Rarg1, Rout
		ll := larch64.ALLV
		sc := larch64.ASCV
		if v.Op == ssa.OpLARCH64LoweredAtomicAdd32 {
			ll = larch64.ALL
			sc = larch64.ASC
		}
		s.Prog(larch64.ASYNC)
		p := s.Prog(ll)
		p.From.Type = obj.TYPE_MEM
		p.From.Reg = v.Args[0].Reg()
		p.To.Type = obj.TYPE_REG
		p.To.Reg = v.Reg0()
		p1 := s.Prog(larch64.AADDVU)
		p1.From.Type = obj.TYPE_REG
		p1.From.Reg = v.Args[1].Reg()
		p1.Reg = v.Reg0()
		p1.To.Type = obj.TYPE_REG
		p1.To.Reg = larch64.REGTMP
		p2 := s.Prog(sc)
		p2.From.Type = obj.TYPE_REG
		p2.From.Reg = larch64.REGTMP
		p2.To.Type = obj.TYPE_MEM
		p2.To.Reg = v.Args[0].Reg()
		p3 := s.Prog(larch64.ABEQ)
		p3.From.Type = obj.TYPE_REG
		p3.From.Reg = larch64.REGTMP
		p3.To.Type = obj.TYPE_BRANCH
		gc.Patch(p3, p)
		s.Prog(larch64.ASYNC)
		p4 := s.Prog(larch64.AADDVU)
		p4.From.Type = obj.TYPE_REG
		p4.From.Reg = v.Args[1].Reg()
		p4.Reg = v.Reg0()
		p4.To.Type = obj.TYPE_REG
		p4.To.Reg = v.Reg0()
	case ssa.OpLARCH64LoweredAtomicAddconst32, ssa.OpLARCH64LoweredAtomicAddconst64:
		// SYNC
		// LL	(Rarg0), Rout
		// ADDV $auxint, Rout, Rtmp
		// SC	Rtmp, (Rarg0)
		// BEQ	Rtmp, -3(PC)
		// SYNC
		// ADDV $auxint, Rout
		ll := larch64.ALLV
		sc := larch64.ASCV
		if v.Op == ssa.OpLARCH64LoweredAtomicAddconst32 {
			ll = larch64.ALL
			sc = larch64.ASC
		}
		s.Prog(larch64.ASYNC)
		p := s.Prog(ll)
		p.From.Type = obj.TYPE_MEM
		p.From.Reg = v.Args[0].Reg()
		p.To.Type = obj.TYPE_REG
		p.To.Reg = v.Reg0()
		p1 := s.Prog(larch64.AADDVU)
		p1.From.Type = obj.TYPE_CONST
		p1.From.Offset = v.AuxInt
		p1.Reg = v.Reg0()
		p1.To.Type = obj.TYPE_REG
		p1.To.Reg = larch64.REGTMP
		p2 := s.Prog(sc)
		p2.From.Type = obj.TYPE_REG
		p2.From.Reg = larch64.REGTMP
		p2.To.Type = obj.TYPE_MEM
		p2.To.Reg = v.Args[0].Reg()
		p3 := s.Prog(larch64.ABEQ)
		p3.From.Type = obj.TYPE_REG
		p3.From.Reg = larch64.REGTMP
		p3.To.Type = obj.TYPE_BRANCH
		gc.Patch(p3, p)
		s.Prog(larch64.ASYNC)
		p4 := s.Prog(larch64.AADDVU)
		p4.From.Type = obj.TYPE_CONST
		p4.From.Offset = v.AuxInt
		p4.Reg = v.Reg0()
		p4.To.Type = obj.TYPE_REG
		p4.To.Reg = v.Reg0()
	case ssa.OpLARCH64LoweredAtomicCas32, ssa.OpLARCH64LoweredAtomicCas64:
		// MOVV $0, Rout
		// SYNC
		// LL	(Rarg0), Rtmp
		// BNE	Rtmp, Rarg1, 4(PC)
		// MOVV Rarg2, Rout
		// SC	Rout, (Rarg0)
		// BEQ	Rout, -4(PC)
		// SYNC
		ll := larch64.ALLV
		sc := larch64.ASCV
		if v.Op == ssa.OpLARCH64LoweredAtomicCas32 {
			ll = larch64.ALL
			sc = larch64.ASC
		}
		p := s.Prog(larch64.AMOVV)
		p.From.Type = obj.TYPE_REG
		p.From.Reg = larch64.REGZERO
		p.To.Type = obj.TYPE_REG
		p.To.Reg = v.Reg0()
		s.Prog(larch64.ASYNC)
		p1 := s.Prog(ll)
		p1.From.Type = obj.TYPE_MEM
		p1.From.Reg = v.Args[0].Reg()
		p1.To.Type = obj.TYPE_REG
		p1.To.Reg = larch64.REGTMP
		p2 := s.Prog(larch64.ABNE)
		p2.From.Type = obj.TYPE_REG
		p2.From.Reg = v.Args[1].Reg()
		p2.Reg = larch64.REGTMP
		p2.To.Type = obj.TYPE_BRANCH
		p3 := s.Prog(larch64.AMOVV)
		p3.From.Type = obj.TYPE_REG
		p3.From.Reg = v.Args[2].Reg()
		p3.To.Type = obj.TYPE_REG
		p3.To.Reg = v.Reg0()
		p4 := s.Prog(sc)
		p4.From.Type = obj.TYPE_REG
		p4.From.Reg = v.Reg0()
		p4.To.Type = obj.TYPE_MEM
		p4.To.Reg = v.Args[0].Reg()
		p5 := s.Prog(larch64.ABEQ)
		p5.From.Type = obj.TYPE_REG
		p5.From.Reg = v.Reg0()
		p5.To.Type = obj.TYPE_BRANCH
		gc.Patch(p5, p1)
		p6 := s.Prog(larch64.ASYNC)
		gc.Patch(p2, p6)
	case ssa.OpLARCH64LoweredNilCheck:
		// Issue a load which will fault if arg is nil.
		p := s.Prog(larch64.AMOVB)
		p.From.Type = obj.TYPE_MEM
		p.From.Reg = v.Args[0].Reg()
		gc.AddAux(&p.From, v)
		p.To.Type = obj.TYPE_REG
		p.To.Reg = larch64.REGTMP
		if logopt.Enabled() {
			logopt.LogOpt(v.Pos, "nilcheck", "genssa", v.Block.Func.Name)
		}
		if gc.Debug_checknil != 0 && v.Pos.Line() > 1 { // v.Pos.Line()==1 in generated wrappers
			gc.Warnl(v.Pos, "generated nil check")
		}
	case ssa.OpLARCH64FPFlagTrue,
		ssa.OpLARCH64FPFlagFalse:
		// MOVV	$0, r
		// BFPF	2(PC)
		// MOVV	$1, r
		branch := larch64.ABFPF
		if v.Op == ssa.OpLARCH64FPFlagFalse {
			branch = larch64.ABFPT
		}
		p := s.Prog(larch64.AMOVV)
		p.From.Type = obj.TYPE_REG
		p.From.Reg = larch64.REGZERO
		p.To.Type = obj.TYPE_REG
		p.To.Reg = v.Reg()
		p2 := s.Prog(branch)
		p2.To.Type = obj.TYPE_BRANCH
		p3 := s.Prog(larch64.AMOVV)
		p3.From.Type = obj.TYPE_CONST
		p3.From.Offset = 1
		p3.To.Type = obj.TYPE_REG
		p3.To.Reg = v.Reg()
		p4 := s.Prog(obj.ANOP) // not a machine instruction, for branch to land
		gc.Patch(p2, p4)
	case ssa.OpLARCH64LoweredGetClosurePtr:
		// Closure pointer is R22 (larch64.REGCTXT).
		gc.CheckLoweredGetClosurePtr(v)
	case ssa.OpLARCH64LoweredGetCallerSP:
		// caller's SP is FixedFrameSize below the address of the first arg
		p := s.Prog(larch64.AMOVV)
		p.From.Type = obj.TYPE_ADDR
		p.From.Offset = -gc.Ctxt.FixedFrameSize()
		p.From.Name = obj.NAME_PARAM
		p.To.Type = obj.TYPE_REG
		p.To.Reg = v.Reg()
	case ssa.OpLARCH64LoweredGetCallerPC:
		p := s.Prog(obj.AGETCALLERPC)
		p.To.Type = obj.TYPE_REG
		p.To.Reg = v.Reg()
	case ssa.OpClobber:
		// TODO: implement for clobberdead experiment. Nop is ok for now.
	default:
		v.Fatalf("genValue not implemented: %s", v.LongString())
	}
}

var blockJump = map[ssa.BlockKind]struct {
	asm, invasm obj.As
}{
	ssa.BlockLARCH64EQ:  {larch64.ABEQ, larch64.ABNE},
	ssa.BlockLARCH64NE:  {larch64.ABNE, larch64.ABEQ},
	ssa.BlockLARCH64LTZ: {larch64.ABLTZ, larch64.ABGEZ},
	ssa.BlockLARCH64GEZ: {larch64.ABGEZ, larch64.ABLTZ},
	ssa.BlockLARCH64LEZ: {larch64.ABLEZ, larch64.ABGTZ},
	ssa.BlockLARCH64GTZ: {larch64.ABGTZ, larch64.ABLEZ},
	ssa.BlockLARCH64FPT: {larch64.ABFPT, larch64.ABFPF},
	ssa.BlockLARCH64FPF: {larch64.ABFPF, larch64.ABFPT},
}

func ssaGenBlock(s *gc.SSAGenState, b, next *ssa.Block) {
	switch b.Kind {
	case ssa.BlockPlain:
		if b.Succs[0].Block() != next {
			p := s.Prog(obj.AJMP)
			p.To.Type = obj.TYPE_BRANCH
			s.Branches = append(s.Branches, gc.Branch{P: p, B: b.Succs[0].Block()})
		}
	case ssa.BlockDefer:
		// defer returns in R1:
		// 0 if we should continue executing
		// 1 if we should jump to deferreturn call
		p := s.Prog(larch64.ABNE)
		p.From.Type = obj.TYPE_REG
		p.From.Reg = larch64.REGZERO
		p.Reg = larch64.REG_R19
		p.To.Type = obj.TYPE_BRANCH
		s.Branches = append(s.Branches, gc.Branch{P: p, B: b.Succs[1].Block()})
		if b.Succs[0].Block() != next {
			p := s.Prog(obj.AJMP)
			p.To.Type = obj.TYPE_BRANCH
			s.Branches = append(s.Branches, gc.Branch{P: p, B: b.Succs[0].Block()})
		}
	case ssa.BlockExit:
	case ssa.BlockRet:
		s.Prog(obj.ARET)
	case ssa.BlockRetJmp:
		p := s.Prog(obj.ARET)
		p.To.Type = obj.TYPE_MEM
		p.To.Name = obj.NAME_EXTERN
		p.To.Sym = b.Aux.(*obj.LSym)
	case ssa.BlockLARCH64EQ, ssa.BlockLARCH64NE,
		ssa.BlockLARCH64LTZ, ssa.BlockLARCH64GEZ,
		ssa.BlockLARCH64LEZ, ssa.BlockLARCH64GTZ,
		ssa.BlockLARCH64FPT, ssa.BlockLARCH64FPF:
		jmp := blockJump[b.Kind]
		var p *obj.Prog
		switch next {
		case b.Succs[0].Block():
			p = s.Br(jmp.invasm, b.Succs[1].Block())
		case b.Succs[1].Block():
			p = s.Br(jmp.asm, b.Succs[0].Block())
		default:
			if b.Likely != ssa.BranchUnlikely {
				p = s.Br(jmp.asm, b.Succs[0].Block())
				s.Br(obj.AJMP, b.Succs[1].Block())
			} else {
				p = s.Br(jmp.invasm, b.Succs[1].Block())
				s.Br(obj.AJMP, b.Succs[0].Block())
			}
		}
		if !b.Controls[0].Type.IsFlags() {
			p.From.Type = obj.TYPE_REG
			p.From.Reg = b.Controls[0].Reg()
		}
	default:
		b.Fatalf("branch not implemented: %s", b.LongString())
	}
}
